(define NFS-PORT 2049)
(define NFS-MAXDATA 8192)
(define NFS-MAXPATHLEN 1024)
(define NFS-MAXNAMLEN 255)
(define NFS-FHSIZE 32)
(define NFS-COOKIESIZE 4)
(define NFS-FIFO-DEV -1)
(define NFSMODE-FMT 61440)
(define NFSMODE-DIR 16384)
(define NFSMODE-CHR 8192)
(define NFSMODE-BLK 24576)
(define NFSMODE-REG 32768)
(define NFSMODE-LNK 40960)
(define NFSMODE-SOCK 49152)
(define NFSMODE-FIFO 4096)
(define char xdr-integer)
(define nfsstat
  (make-xdr-enumeration
    'nfsstat
    `((NFS-OK . 0)
      (NFSERR-PERM . 1)
      (NFSERR-NOENT . 2)
      (NFSERR-IO . 5)
      (NFSERR-NXIO . 6)
      (NFSERR-ACCES . 13)
      (NFSERR-EXIST . 17)
      (NFSERR-NODEV . 19)
      (NFSERR-NOTDIR . 20)
      (NFSERR-ISDIR . 21)
      (NFSERR-FBIG . 27)
      (NFSERR-NOSPC . 28)
      (NFSERR-ROFS . 30)
      (NFSERR-NAMETOOLONG . 63)
      (NFSERR-NOTEMPTY . 66)
      (NFSERR-DQUOT . 69)
      (NFSERR-STALE . 70)
      (NFSERR-WFLUSH . 99))))
(define ftype
  (make-xdr-enumeration
    'ftype
    `((NFNON . 0)
      (NFREG . 1)
      (NFDIR . 2)
      (NFBLK . 3)
      (NFCHR . 4)
      (NFLNK . 5)
      (NFSOCK . 6)
      (NFBAD . 7)
      (NFFIFO . 8))))
(define nfs-fh
  (make-xdr-struct-type
    (list (make-xdr-fixed-length-opaque-array NFS-FHSIZE))))
(define nfstime
  (make-xdr-struct-type
    (list xdr-unsigned-integer xdr-unsigned-integer)))
(define fattr
  (make-xdr-struct-type
    (list ftype
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          nfstime
          nfstime
          nfstime)))
(define sattr
  (make-xdr-struct-type
    (list xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          nfstime
          nfstime)))
(define filename (make-xdr-string NFS-MAXNAMLEN))
(define nfspath (make-xdr-string NFS-MAXPATHLEN))
(define attrstat
  (make-xdr-union-type
    nfsstat
    `((NFS-OK unquote fattr))
    xdr-void))
(define sattrargs
  (make-xdr-struct-type (list nfs-fh sattr)))
(define diropargs
  (make-xdr-struct-type (list nfs-fh filename)))
(define diropokres
  (make-xdr-struct-type (list nfs-fh fattr)))
(define diropres
  (make-xdr-union-type
    nfsstat
    `((NFS-OK unquote diropokres))
    xdr-void))
(define readlinkres
  (make-xdr-union-type
    nfsstat
    `((NFS-OK unquote nfspath))
    xdr-void))
(define readargs
  (make-xdr-struct-type
    (list nfs-fh
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer)))
(define readokres
  (make-xdr-struct-type
    (list fattr
          (make-xdr-variable-length-opaque-array
            NFS-MAXDATA))))
(define readres
  (make-xdr-union-type
    nfsstat
    `((NFS-OK unquote readokres))
    xdr-void))
(define writeargs
  (make-xdr-struct-type
    (list nfs-fh
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          (make-xdr-variable-length-opaque-array
            NFS-MAXDATA))))
(define createargs
  (make-xdr-struct-type (list diropargs sattr)))
(define renameargs
  (make-xdr-struct-type (list diropargs diropargs)))
(define linkargs
  (make-xdr-struct-type (list nfs-fh diropargs)))
(define symlinkargs
  (make-xdr-struct-type
    (list diropargs nfspath sattr)))
(define nfscookie
  (make-xdr-fixed-length-opaque-array
    NFS-COOKIESIZE))
(define readdirargs
  (make-xdr-struct-type
    (list nfs-fh nfscookie xdr-unsigned-integer)))
(define entry
  (make-xdr-struct-type
    (list xdr-unsigned-integer
          filename
          nfscookie
          (make-xdr-union-type
            xdr-boolean
            `((TRUE unquote (lambda () entry))
              (FALSE unquote xdr-void))
            #f))))
(define dirlist
  (make-xdr-struct-type
    (list (make-xdr-union-type
            xdr-boolean
            `((TRUE unquote entry) (FALSE unquote xdr-void))
            #f)
          xdr-boolean)))
(define readdirres
  (make-xdr-union-type
    nfsstat
    `((NFS-OK unquote dirlist))
    xdr-void))
(define statfsokres
  (make-xdr-struct-type
    (list xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer
          xdr-unsigned-integer)))
(define statfsres
  (make-xdr-union-type
    nfsstat
    `((NFS-OK unquote statfsokres))
    xdr-void))
(define (make-NFS-PROGRAM-server versions+procs)
  (define %program-versions
    `(("NFS_VERSION"
       2
       ("NFSPROC_NULL" 0 ,xdr-void ())
       ("NFSPROC_GETATTR" 1 ,attrstat (,nfs-fh))
       ("NFSPROC_SETATTR" 2 ,attrstat (,sattrargs))
       ("NFSPROC_ROOT" 3 ,xdr-void ())
       ("NFSPROC_LOOKUP" 4 ,diropres (,diropargs))
       ("NFSPROC_READLINK" 5 ,readlinkres (,nfs-fh))
       ("NFSPROC_READ" 6 ,readres (,readargs))
       ("NFSPROC_WRITECACHE" 7 ,xdr-void ())
       ("NFSPROC_WRITE" 8 ,attrstat (,writeargs))
       ("NFSPROC_CREATE" 9 ,diropres (,createargs))
       ("NFSPROC_REMOVE" 10 ,nfsstat (,diropargs))
       ("NFSPROC_RENAME" 11 ,nfsstat (,renameargs))
       ("NFSPROC_LINK" 12 ,nfsstat (,linkargs))
       ("NFSPROC_SYMLINK" 13 ,nfsstat (,symlinkargs))
       ("NFSPROC_MKDIR" 14 ,diropres (,createargs))
       ("NFSPROC_RMDIR" 15 ,nfsstat (,diropargs))
       ("NFSPROC_READDIR" 16 ,readdirres (,readdirargs))
       ("NFSPROC_STATFS" 17 ,statfsres (,nfs-fh)))))
  (make-rpc-program
    100003
    (filter-map
      (lambda (version+procs)
        (let* ((name (car version+procs))
               (version (assoc name %program-versions)))
          (and version
               (let ((number (car (cdr version)))
                     (user-procs (cdr version+procs)))
                 (make-rpc-program-version
                   number
                   (let ((version-procs (cddr version)))
                     (filter-map
                       (lambda (user-proc)
                         (let* ((name (car user-proc))
                                (handler (cdr user-proc))
                                (proc (assoc name version-procs)))
                           (and proc
                                (let ((number (car (cdr proc)))
                                      (ret-type (cadr (cdr proc)))
                                      (arg-types (caddr (cdr proc))))
                                  (make-rpc-procedure
                                    number
                                    (cond ((null? arg-types) xdr-void)
                                          ((null? (cdr arg-types))
                                           (car arg-types))
                                          (else
                                           (apply make-xdr-struct-type
                                                  arg-types)))
                                    ret-type
                                    handler)))))
                       user-procs)))))))
      versions+procs)))
